/* ***************************************************



***************************************************** */

#include <LEDA/graph/graph.h>
#include <LEDA/graph/node_array.h>
#include <LEDA/core/list.h>
#include <LEDA/core/array.h>
#include <LEDA/core/string.h>
#include <LEDA/graphics/graphwin.h>
#include <LEDA/numbers/integer.h>
#include <time.h>
#include <iostream>

using namespace leda;
using namespace std;



void triangulation(graph &G,graph &H,list< list<node> > &F,list<edge> &le);
void operation(graph &G,graph &H,node v,list< list<node> > &F,list<edge> &le);
int edge_exist(node v1,node v2);
void c_component(graph &G,graph &H,node_array<int> &node_ord,list<node> &C,list<node> &L);
void add(graph &G,graph &H,list< list<node> > &F,list<node> &L,list<node> &L2,list<edge> &le);
void N_G(list<node> &L,list<node> &S);


int main()
{
  graph G,H;
  G.read("graph.gw");
  H=G;

  G.make_undirected();
  H.make_undirected();

  list< list<node> > F;
  list<edge> le;

  //H=(V,E+F) is a minimal triangulation of G=(V,E)
  
  F.clear();
  le.clear();

  triangulation(G,H,F,le);

  H.write("out.gw"); 

}


/* ************************************************************
   triangulation F $B$r:n@.(B
*************************************************************** */

void triangulation(graph &G,graph &H,list< list<node> > &F,list<edge> &le)
{
  node v;

  F.clear();
  le.clear();
  int n = 0;
  forall_nodes(v,H)
      operation(G,H,v,F,le);
    


}

void operation(graph &G,graph &H,node v,list< list<node> > &F,list<edge> &le)
{
  list<node> C,L,L2;
  node_array<int> node_ord(H,1);
  node_ord[v]=0;
  node v1;
  //C_G(N_H[v])$B$O(Bnode_ord$B$K(B1$B$r$D$1$k(B
  forall_adj_nodes(v1,v)
    node_ord[v1]=0;

  node w;
  //C_G(N_H[v])$B$r(Blist C$B$KJ]B8$9$k(B
  forall_nodes(w,H)
    if(node_ord[w]==1)
      C.append(w);
  while( !C.empty() ){
    c_component(G,H,node_ord,C,L);
    N_G(L,L2);
    if(L2.size() > 1)
      add(G,H,F,L,L2,le);
    L.clear();
    L2.clear();
  }

}

//v1,v2$B4V$K(Bedge$B$,B8:_$7$J$$$+$rH=Dj(B
int edge_exist(node v1,node v2)
{
  node w;
  if(v1 == v2)
    return 0;

  forall_adj_nodes(w,v1)
    if(w == v2)
      return 0;

  return 1;
}

//connected component C in C_G(N_H[v])
void c_component(graph &G,graph &H,node_array<int> &node_ord,list<node> &C,list<node> &L)
{
  list_item it = C.first();

  node v = C[it];
  node_ord[v] = 0;
  L.append(v);
  C.erase(it);

  node w;
  forall_adj_nodes(w,v)
    if( node_ord[w]==1 )
      c_component(G,H,node_ord,C,L);
}


void add(graph &G,graph &H,list< list<node> > &F,list<node> &L,list<node> &L2,list<edge> &le)
{
  list_item it = L2.first();
  list_item it_succ;
  while( it != L2.last() ){
    it_succ = L2.succ(it);
    while( it_succ != nil ){
      if( edge_exist(L2[it],L2[it_succ]) ){
	edge e = H.new_edge(L2[it],L2[it_succ]);
	le.append(e);
	list<node> f1;
	f1.append(L2[it]);
	f1.append(L2[it_succ]);
	F.append(f1);
      }
      it_succ = L2.succ(it_succ);
    }
    it = L.succ(it);
  }
}

//N_G(C)$B$r(BS$B$KJ]B8$9$k(B
void N_G(list<node> &L,list<node> &S)
{
  node v,w;
  forall(v,L){
    forall_adj_nodes(w,v){
      if(S.search(w)==nil)
	S.append(w);
    }
  }
  forall(v,L)
    S.remove(v);
}
